#
# Infrared object database
#
# This is an experimental sqlite database format for recording,
# classifying, and managing IR Symbols. It's not ready for
# any kind of Prime Time yet, but it's better than snatches of
# text files.
# We make no attempt to hide the sqlite-ness of the database;
# this is really just a convenient schema and a thin shell around
# the sqlite API.
# Actual Symbols are stored in pickled form.
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sqlite3
import cPickle
import os
import ir


#
# Which database version do we want to work with?
#
VERSION=3


#
# We need to pickle all subclasses of ir.Symbol. (Fie!)
#
def s2b(symbol):
	return buffer(cPickle.dumps(symbol, 2))
def b2s(blob):
	return cPickle.loads(blob)

def adapt(cls, action):
	sqlite3.register_adapter(cls, action)
	for sub in cls.__subclasses__():
		adapt(sub, action)

adapt(ir.Symbol, s2b)
sqlite3.register_converter("PICKLED", b2s)


#
# An IR database object. This is a translucent wrapper around an sqlite3 connection.
#
class IRDb(object):
	""" Infrared database object.

		IRDb is a translucent wrapper around sqlite3.connection.
		It takes care of storing several ir.* classes in the database (using pickling).
		It also provides for matching against Symbol sets, using queries to cut down
		the competition and then using fuzzy linear scans to finish.
	"""
	def __init__(self, path=None, **kwargs):
		self._con = None
		if path:
			self.open(path, **kwargs)

	def open(self, path, **kwargs):
		self.path = path
		self._con = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES, **kwargs)
		self._con.row_factory = sqlite3.Row
		version = None
		try:
			row = self.evaluate(
				'''SELECT version, creator, description FROM config;''')
			if row is None:
				if not os.stat(path).st_size:	# empty file; sqlite just made it
					os.remove(path)				# so ditch it
				raise sqlite3.OperationalError("Database empty or corrupt")
			self.version = row[0]
			self.creator = row[1]
			self.description = row[2]
		except sqlite3.OperationalError:
			pass
		if self.version != VERSION:
			raise sqlite3.OperationalError("Incompatible database version (got %d need %d)" % (version, VERSION))

	def close(self):
		self._con.close()
		self._con = None

	def commit(self):
		self._con.commit()

	def abort(self):
		self._con.abort()

	def find_symbol(self, group, name=None):
		""" Retrieve a symbol (only) by group and name. """
		if '/' in group:
			(group, name) = group.split('/')
		elif name is not None and '/' in name:
			(group, name) = name.split('/')
		if group is None or name is None:
			return None
		return self.evaluate1(
			'''SELECT symbol FROM symbol WHERE grouping=? and name=?;''',
			(group, name))

	def match_code(self, code):
		""" Match a Code against Symbols in the database. Return a matching row, if any. """
		assert isinstance(code, ir.Code)
		match = []
		if code.type().exactMatch:
			search = self.execute('''SELECT grouping, name, symbol FROM symbol WHERE symbol=?;''',
				(code.symbol(), ))
		else:
			search = self.execute('''SELECT grouping, name, symbol FROM symbol WHERE type=? and carrier=?;''',
				(code.typename(), code.carrier))
		for row in search:
			if row['symbol'].matches(code):
				return row
		return None

	def store(self, symbol, id):
		""" Store a Symbol in the database. """
		assert isinstance(symbol, ir.Symbol)
		(group, name) = id
		self.execute('''INSERT OR REPLACE INTO symbol (
			grouping, name, symbol, type, carrier
			) VALUES (?, ?, ?, ?, ?);''', (
			group, name, symbol, symbol.typename(), symbol.carrier
		))


	def execute(self, command, args=()):
		return self._con.execute(command, args)

	def script(self, script):
		return self._con.executescript(script)

	def evaluate(self, command, args=()):
		""" Get a single value by SQL. Returns None if nothing is found. """
		return self.execute(command, args).fetchone()

	def evaluate1(self, command, args=()):
		row = self.evaluate(command, args)
		return row[0] if row else None


	def load_from(self, source_path):
		db = IRDb(source_path)
		cursor = db.execute('''SELECT grouping, name, symbol FROM symbol;''')
		for row in cursor:
			self.store(row['symbol'], (row['grouping'], row['name']))
		db.close()
