#
# Global Cache GC-100/iTach or similar - Network device interface
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import deque
import re
import sys
import socket
import time
from distutils.version import StrictVersion

import asyn
import asyn.resolve
import asyn.utility

from core import Error
from core import DEVICE_GC100, DEVICE_ITACH, DEVICE_FLEX, SERIAL_BASE
import gcir
import gcdev

DEBUG = None


#
# Device handling heuristics
#
PROBE_DELAY = 0.15

V_GC100_GET_CMDS = StrictVersion("3.0")			# min version for get_{IR,SERIAL} commands


#
# A GC-style network device.
#
class GCNet(asyn.Callable, asyn.utility.Idler):
	""" a GC-100, iTach, or similar network device with attached hardware device(s).

		Global Cache network devices uses a modular model where each network device
		contains one or more device modules. GCNet represents the network interface
		itself; and each device subfunction gets its own instance of Device.
	"""

	input_scan = asyn.scan.Regex([
		(r'unknowncommand (\d+)\r', "error"),	# GC-100 style
		(r'ERR_\d+:\d+,(\d+)\r', "error"),		# classic iTach style
		(r'ERR( \w+)\r', "error"),				# iTach flex style (keep space to distinguish)
		(r'device,(.*?)\r', "device"),
		(r'endlistdevices\r', "devices done"),
		(r'version,\d+,(.*?)\r', "version"),	# GC-100 version
		(r'(\d{3}-\d{4}-\d{2})\r', "version"),	# iTach version (both kinds)
		(r'IR,(.*?)\r', "IR"),
		(r'SERIAL,(.*?)\r', "SERIAL"),
		(r'devicechange\r', 'devicechange'),
		(r'IR Learner Enabled\r', "start learning"),
		(r'IR Learner Disabled\r', "stop learning"),
		(r'sendir,(.*?)\r', "learned"), # actually ends in \r\n...
		(r'\n', None),					# ... so eat that
		(r'(.*?)\r', "reply"),			# anything else
	])

	def __init__(self, control, callout=None):
		""" Create a new network device interface given an asyn Controller.

			The object will be initially inert. Call open() to start discovery.
		"""
		asyn.Callable.__init__(self, callout=callout)
		asyn.utility.Idler.__init__(self, control)
		self.control = control							# our controller
		self.hostdev = None								# no host (network) device yet

	def open(self, socket):
		""" Open a GCNet for business. The socket must be a connected stream socket.

			This starts auto-discovery of the device, which will determine
			its hardware type, basic function, and device list.
			This will take a little while; when it's done, you'll
			either get upcalled with the device dictionary, or an error.
			The class is basically not functional until then.
		"""
		self.hostdev = asyn.selectable.Stream(self.control, socket, callout=self._calldown)
		self.hostdev.scan = GCNet.input_scan
		self.ipaddr = socket.getpeername()[0]
		self.type = None								# device type unknown as yet
		self.can_learn = False							# no learning mode support
		self.can_subquery = True						# has {get,set}_{NET,IR,SERIAL} commands
		self._learning = False							# learning mode (if any) is off
		self.network = "ETHERNET"						# GC-100 default (iTach can discover this)
		self._errors = None								# error code-to-string mapping
		self.version = None								# board version
		self.devmap = None								# device map (none yet)
		self._devmap = { }								# (start collecting internally)
		self._probe_count = None						# initial discovery marker
		self._irseq = 1									# ir sequence number generator
		self._serial_port = SERIAL_BASE					# assign sequential serial ports
		self.command('mumblefrotz')						# intentionally bad command to triage device type

	def close(self):
		""" Shut down this GCNet. """
		self.idle_cancel()
		if self.devmap:
			for dev in self.devmap.values():
				dev.close()
		if self.hostdev:
			self.hostdev.close()
			self.hostdev = None

	def idle(self):
		""" Issue a no-op command to check connectivity. """
		self.command("getversion")						# (pointlessly) re-query board version

	def error_message(self, code):
		""" Translate an "unknowncommand" error code into text. """
		return self._errors[code] if code in self._errors else 'error %d' % code

	def command(self, cmd):
		""" Send a raw (complete) command to the device. """
		if DEBUG: DEBUG(self, '->', cmd)
		self.hostdev.write(cmd + '\r')


	#
	# Inquiries about contained devices
	#
	def all(self):
		""" Iterate over all devices on this GCNet. """
		return self.devmap.itervalues() if self.devmap else []

	def all_type(self, devclass=None):
		""" Iterate over all devices on this GCNet of a given device type. """
		if self.devmap:
			for dev in self.devmap.itervalues():
				if devclass is None or isinstance(dev, devclass):
					yield dev


	#
	# iTach devices (but not GC-100s) include a dedicated (very) short-range IR receiver
	# on each interface. This is a function of the iTach itself, not of any of its
	# device modules, and is basically stuck on top of the regular software stack.
	# Be sure to hold your remote very closely (<1") and aim at the little hole precisely.
	#
	def start_learning(self):
		""" Tell an iTach to enter learning mode.

			Once learning mode is active, the GCNet will upcall ir.Events until learning
			mode is cancelled. Make sure your callback is prepared.
		"""
		if self.can_learn:
			self.command("get_IRL")
		else:
			self.callout_error(Error("device has no learning mode"))

	def stop_learning(self):
		""" Terminate learning mode. """
		if self.can_learn:
			self.command("stop_IRL")


	#
	# Unified input callback.
	#
	def _calldown(self, ctx, data=None):
		""" Unified GCNet callback driver. """
		if ctx.error:
			return self.callout(ctx)
		if ctx.state == 'CLOSE':			# connection lost
			self._io = None					# any retry to be done by upstream
			return self.callout('close')	# ... so let them know
		if ctx.state == 'END':
			return self.callout(ctx)
		if DEBUG: DEBUG(self, '<-', ctx, data)
		if ctx.state == 'error':
			if self.type is None:						# this one was on purpose
				if data == "14":
					self.type = DEVICE_GC100			# it's a GC-100 of some kind
					self._errors = ERRORS_GC_100
					self._irqueue = deque()				# unified IR queue (one per netdev)
				elif data == "001":
					self.type = DEVICE_ITACH			# it's a discrete iTach of some kind
					self._errors = ERRORS_ITACH
					self._irqueue = None				# per-device IR queue
					self.can_learn = True				# supports learning mode
				elif data == ' 001':
					self.type = DEVICE_FLEX				# it's an iTach flex
					self._errors = ERRORS_FLEX
					self._irqueue = None				# per-device IR queue
					self.can_learn = True				# supports learning mode
				else:
					# neither GC-100 nor iTach. Fill in blanks here...
					self.callout(Error("Unknown GC-100-like device (returned error %s)" % data))
					return
				self.command('getversion')				# get board version
				self._probe_count = 1					# start counting devices in flight
				self.command('getdevices')				# start device enumeration
			else:
				# Genuine error. Report upstream.
				code = data.lstrip() if self.type == DEVICE_FLEX else int(data)
				self.callout_error(Error(code, self._errors[code]))
		elif ctx.state == 'device':						# device enumeration ("2,3 IR")
			dev, _, r = data.partition(',')
			cnt, _, type = r.partition(' ')
			if dev == "0":								# iTach network device
				self.network = type
			for addr in range(1, int(cnt)+1):
				self._probe_count += 1					# record device probe in flight
				dev_addr = "%s:%d" % (dev, addr)
				if type == "RELAY":						# no sub-query for relays
					self._create_dev(gcdev.Relay, dev_addr, None)
				elif self.can_subquery:
					time.sleep(PROBE_DELAY)				# should use a queue, but we're lazy
					if type.startswith("IR"):			# IR, IR_BLASTER, IRTRIPORT, ...
						type = "IR"
					self.command("get_%s,%s" % (type, dev_addr)) # issue specific probe
				elif type == 'SERIAL':
					self._create_dev(gcdev.Serial, dev_addr, '')
				else:	# IR
					self._create_dev(gcdev.IREmitter, dev_addr, "IR assumed") # configure as IR emitter
		elif ctx.state == 'IR':							# 2:1,function
			addr, type = data.split(',', 1)
			if type in ["SENSOR", "SENSOR_NOTIFY"]:
				self._create_dev(gcdev.Sensor, addr, type) # configured for input
			elif type == "LED_LIGHTING":
				self._create_dev(gcdev.LEDLight, addr, type) # configured for LED dimmer driving
			elif type.startswith("IRTRIPORT"):			# single report but represents three devices
				main, sub = addr.split(':')
				assert sub == "1"
				self._probe_count += 2					# we're turning one reported device into three
				self._create_dev(gcdev.IREmitter, "%s:1" % main, "IRTRIPORT")
				self._create_dev(gcdev.IREmitter, "%s:2" % main, "IRTRIPORT")
				self._create_dev(gcdev.IREmitter, "%s:3" % main, type)
			else:
				self._create_dev(gcdev.IREmitter, addr, type) # configured as IR emitter
		elif ctx.state == 'SERIAL':						# 2:1,serial parameters
			addr, type = data.split(',', 1)
			self._create_dev(gcdev.Serial, addr, type)
		elif ctx.state == 'devices done':				# device list complete
			self._finish()								# ready to wrap this up
		elif ctx.state == 'version':
			if self.version is None:
				self.version = data							# board version
				if self.type == DEVICE_GC100 and V_GC100_GET_CMDS > self.version.partition('-')[0]:
					self.can_subquery = False				# too old
		elif ctx.state == 'start learning':				# enter learning mode
			self._learning = True
			self.callout('learning', True)
		elif ctx.state == 'stop learning':				# exit learning mode
			self._learning = False
			self.callout('learning', False)
		elif ctx.state == 'learned':					# IR signal learned
			self.callout('learned', self._iTach_learned_event(data))
		elif ctx.state == 'devicechange':				# probably defunct
			pass
		elif ctx.state == 'reply':						# any other reply
			try:
				(cmd, addr, args) = data.split(',', 2)
				if self.devmap and addr in self.devmap:
					self.devmap[addr]._calldown(ctx, cmd, args)
				else:
					self.callout(Error("Unconsumed event", data))
			except ValueError:
				self.callout(Error("Unrecognized event", data))
		self.idle_activity()

	# iTach learned data is in a weird command-to-virtual-device form. Just peel off the gunk
	def _iTach_learned_event(self, data):
		m = re.match(r'\d:\d,\d+,(\d+),1,1,(.*)', data)
		if m:
			frequency, pulses = m.group(1, 2)
			frequency = "%d" % round(int(frequency), -3) # match IRL/IRE behavior
			return gcir.Event(frequency + "," + pulses)

	# create a device object and record it
	def _create_dev(self, cls, address, specific):
		""" Internal device creation method. Reply driven. """
		dev = cls(self, address, specific)
		assert not address in self._devmap
		self._devmap[address] = dev
		self._finish()

	def _finish(self):
		self._probe_count -= 1
		if self._probe_count == 0:
			# all discovered devices have reported in. Declare ourselves ready to serve
			self.devmap = self._devmap						# make device map visible
			del self._devmap								# (clean up)
			del self._probe_count
			del self._serial_port
			self.callout('ready', self.devmap)				# report ready

	def __str__(self):
		r = "<%s:%s" % (self.ipaddr, self.type if self.type else "UNKNOWN")
		if self.version:
			r += " %s" % self.version
		if self.devmap:
			r += " %d device(s)" % len(self.devmap)
		return r + ">"

	def __repr__(self): return str(self)


#
# The GC-100 and (various) iTach devices are similar,
# but they have different error code mappings.
# The good news is that this allows us to figure out what we're talking to...
#
_error_command = 'invalid command'
_error_syntax = 'bad command syntax'
_error_module = 'invalid module address'
_error_connector = 'invalid connector address'
_error_module_mismatch = 'command invalid for module type'
_error_ir_to_sensor = 'ir output command to sensor input address'
_error_sensor_to_ir = 'sensor command to ir output address'
_error_ir_id = 'invalid id value'
_error_pulse_count = 'invalid pulse count'
_error_offset = 'invalid ir offset'
_error_odd_ir = 'odd number of ir transitions'
_error_frequency = 'invalid frequency value'
_error_repeat = 'invalid repeat value'
_error_pulse = 'invalid pulse data'
_error_missing_cr = 'missing carriage return'
_error_baud = 'invalid baud rate'
_error_flow = 'invalid flow control setting'
_error_parity = 'invalid parity setting'
_error_locked = 'settings are locked'

ERRORS_GC_100 = {
	1:	'timed out',
	2:	_error_module,
	3:	_error_module,
	4:	_error_connector,
	5:	_error_ir_to_sensor,
	6:	_error_ir_to_sensor,
	7:	_error_ir_to_sensor,
	8:	_error_offset,
	9:	_error_pulse_count,
	10:	_error_odd_ir,
	11:	_error_module_mismatch,
	12:	_error_missing_cr,
	13:	_error_sensor_to_ir,
	14:	_error_command,
	15:	_error_pulse_count,
	16:	_error_odd_ir,
	21:	_error_module_mismatch,
	23:	_error_module_mismatch
}

ERRORS_ITACH = {
	1:	_error_command,
	2:	_error_module,
	3:	_error_connector,
	4:	_error_ir_id,
	5:	_error_frequency,
	6:	_error_repeat,
	7:	_error_offset,
	8:	_error_pulse_count,
	9:	_error_pulse,
	10:	_error_odd_ir,
	11:	_error_missing_cr,
	12:	_error_repeat,
	13:	_error_ir_to_sensor,
	14:	'blaster command to non-blaster connector',
	15:	_error_missing_cr,
	16:	_error_missing_cr,
	17:	_error_syntax,
	18:	_error_sensor_to_ir,
	19:	'repeated ir transmission failure',
	20:	_error_pulse_count,
	21:	'symbol odd boundary',
	22:	'undefined symbol',
	23:	'unknown option',
	24:	_error_baud,
	25:	_error_flow,
	26:	_error_parity,
	27:	_error_locked,
}

ERRORS_FLEX = {
	'001':		_error_command,
	'002':		_error_syntax,
	'003':		_error_connector,
	'004':		_error_missing_cr,
	'005':		_error_module_mismatch,
	'006':		_error_locked,
	'IR001':	_error_ir_id,
	'IR002':	_error_frequency,
	'IR003':	_error_repeat,
	'IR004':	_error_offset,
	'IR005':	_error_pulse,
	'IR006':	_error_odd_ir,
	'IR007':	'too many IR pulses',
	'SL001':	_error_baud,
	'SL002':	_error_flow,
	'SL003':	_error_parity,
}


#
# Ad-hoc test
#
if __name__ == "__main__":

	def cb(ctx, data=None):
		if ctx.state == 'ready':
			print "%s %s %d device(s)" % (net.type, net.version, len(data))
		print ctx, data

	def commands(ctx, cmd=None):
		if ctx.state == 'END':
			return control.close()
		if cmd:
			net.command(cmd)

	host = sys.argv[1]
	control = asyn.Controller()
	net = GCNet(control, host)
	net.add_callout(cb)
	control.commands(commands)
	control.run()
