#
# ir_sony - Sony standard IR codes
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir

name = "Sony"
exactMatch = True


#
# Characteristics of the Sony protocol:
# Sony pulses are (off, on), and a single repeat is actually
#	ON(96), (off,on){12}, PAUSE(1000 or so)
# which we transliterate into
#	LEADER, (on,off){11}, (on=24|48, off=PAUSE)
# to fit our (on, off) form.
#
CARRIER = 40000
LEADER = ir.Pulse(96,24)
ZERO = ir.Pulse(24,24)
ONE = ir.Pulse(48,24)
PAUSE = 1000
PAUSE_SLOP = 200

FUNC_BITS = 7					# bits in function field
DEV_BITS = [13, 5, 8]			# bits in device field (multiple choices, in order of likelihood)


#
# Derived constants
#
ZERO_CUT = ir.Pulse(ZERO.on, PAUSE)
ONE_CUT = ir.Pulse(ONE.on, PAUSE)


class Symbol(ir.Symbol):

	carrier = CARRIER

	def __init__(self, devbits, dev, func):
		ir.Symbol.__init__(self, module)
		self.devbits = devbits
		self.device = dev
		self.function = func

	def matches(self, code):
		return (isinstance(code, Code) and
			self.device == code.device and
			self.function == code.function)

	def code(self, repeats=2):
		return Code(self.devbits, self.device, self.function, repeats)

	#
	# Canonical string form
	#
	def __str__(self):
		return "<%s:%d %04x/%02x>" % (name, self.devbits, self.device, self.function)

	def __repr__(self): return str(self)

	@staticmethod
	def from_text(text):
		(db, rest) = text.split()
		(dev, fun) = rest.split('/')
		return Symbol(int(db), int(dev, 16), int(fun, 16))


class Code(ir.Code):

	carrier = CARRIER

	def __init__(self, devbits, dev, func, repeat):
		ir.Code.__init__(self, module)
		self.devbits = devbits
		self.device = dev
		self.function = func
		self.repeat = repeat

	def __str__(self):
		return "<Code:Sony.%d %x/%02x{%d}>" % (self.devbits, self.device, self.function, self.repeat)

	def __repr__(self): return str(self)

	def symbol(self):
		return Symbol(self.devbits, self.device, self.function)

	def form(self):
		""" Make abstract IR form. Returns (base, repeat, count) tuple. """
		bitsize = self.devbits + FUNC_BITS
		payload = self.device << FUNC_BITS | self.function
		def bits(mask, count):
			for n in range(0, count): yield mask & 1; mask = mask >> 1
		base = [LEADER] + [ONE if x else ZERO for x in bits(payload, bitsize)]
		base[bitsize] = ZERO_CUT if base[bitsize] == ZERO else ONE_CUT # rectify on/off shift and add pause
		return (base, None, self.repeat)


#
# Type check and encode
#
def claim(event):
	""" If event looks like a Sony signal, encode it as our Code. Return None if it doesn't. """

	#
	# A Sony pulsetrain looks like this:
	#	LEADER (( ZERO | ONE ){12} TRAILER )*
	# but bits are (off, on) and thus appear shifted "half left" into the LEADER pulse.
	# See above.
	#
	if event.carrier != CARRIER: return None
	if event[0] != LEADER: return None

	# Extract bits. There is no useful field differentiation.
	pulses = event.pulsetrain
	def bit(p):
		if p == ZERO: return 0
		elif p == ONE: return 1
		else: raise ValueError("invalid bit")

	# try each bit size
	for devbits in DEV_BITS:
		payload_bits = devbits + FUNC_BITS
		total_bits = payload_bits + 1
		if len(pulses) < total_bits:
			continue	# too short
		try:
			code = map(bit, pulses[1:payload_bits]) # all but last bit
		except ValueError:
			continue
		if pulses[payload_bits].__eq__(ZERO_CUT, PAUSE_SLOP):
			code.append(0)
		elif pulses[payload_bits].__eq__(ONE_CUT, PAUSE_SLOP):
			code.append(1)
		else:
			continue
		code.reverse()
		bits = reduce(lambda a, b: a*2+b, code, 0)

		# Check repetition count
		(repeats, left) = divmod(len(pulses), total_bits)
		if left: continue
		for i in range(total_bits, len(pulses), total_bits):
			if pulses[0:total_bits-1] != pulses[i:i+total_bits-1]: return None
			if not pulses[total_bits-1].__eq__(pulses[i+total_bits-1], PAUSE_SLOP): return None

		# all is well; create our Code
		return Code(devbits, bits >> FUNC_BITS, bits & ((1<<FUNC_BITS)-1), repeats)

	# no match for any size
	return None


#
# Fetch our own module reference
#
import sys
module = sys.modules[__name__]
